package org.jboss.seam.cache;

import org.jboss.seam.statistics.StatisticsGenerator;

import java.util.LinkedList;
import java.util.Arrays;
import java.util.List;

/**
 * Statistics for caches.
 * CACHE_HIT and MISS can be used for most caches. The FileSystemCache has a third field BOOST_CACHE_HIT
 * when hitting the in memory cache.
 *
 * @author Sebastian Hennebrueder
 */

public class CacheStatistics implements StatisticsGenerator {
	private Byte semaphore = 1;
	private long total;
	private long boostCacheHit;
	private long cacheHit;
	private String name;

	private String groupname = "org.jboss.seam.cache";

	public CacheStatistics(String name) {
		this.name = name;
	}

	/**
	 * @param name - the cache name, used when printing the report
	 * @param groupname - group name of the statistics
	 */
	public CacheStatistics(String name, String groupname) {
		this.name = name;

		this.groupname=groupname;
	}

	private LinkedList<EntryType> last1000 = new LinkedList<EntryType>();

	public enum EntryType {
		MISS, CACHE_HIT, BOOST_CACHE_HIT
	}

	/**
	 * Add an statistic entry
	 * @param entryType
	 */
	public void addEntry(EntryType entryType) {
		synchronized (semaphore) {
			total++;
			last1000.add(entryType);
			if (last1000.size() > 1000)
				last1000.removeFirst();
			switch (entryType) {
				case MISS:
					break;
				case CACHE_HIT:
					cacheHit++;
          break;
        case BOOST_CACHE_HIT:
					boostCacheHit++;
					break;
			}
		}
	}


	public void reset() {
		synchronized (semaphore) {
			total = 0;
			cacheHit = 0;
			boostCacheHit = 0;
			last1000.clear();
		}
	}


	/**
	 * Total access = Cache hit + Boost Cache hit + Cache miss
	 * @return
	 */
	public long getTotal() {
		return total;
	}

	/**
	 * Total of the last 1000 or lower if you did not have already 1000 requests
	 *
	 * @return
	 */
	public long getLatestTotal() {
		return count(EntryType.BOOST_CACHE_HIT, EntryType.CACHE_HIT, EntryType.MISS);
	}


	/**
	 * Cache hits directly from the in memory cache from the last 1000
	 *
	 * @return
	 */
	public long getLatestBoostCacheHit() {
		return count(EntryType.BOOST_CACHE_HIT);
	}


	/**
	 * Cache hits directly from the in memory cache
	 *
	 * @return
	 */
	public long getBoostCacheHit() {
		return boostCacheHit;
	}

	/**
	 * 	Cache hits from the last 1000
	 * @return
	 */
	public long getCacheHit() {
		return cacheHit;
	}

	/**
	 * Cache hits from the last 1000
	 *
	 * @return
	 */
	public long getLatestCacheHit() {
		return count(EntryType.CACHE_HIT);
	}


	private long count(EntryType... wantedEntryType) {
		List<EntryType> wanted = Arrays.asList(wantedEntryType);
		long count = 0;
		synchronized (semaphore) {
			for (EntryType entryType : last1000) {
				if (wanted.contains(entryType))
					count++;
			}
		}
		return count;
	}

	/**
	 * Create a String containing a report from the hold statistics
	 * @return
	 */
	public String createReport() {
		String totals = String.format("%s statistics Total: %d Cachehits: %d (in %% %.1f) Boost cache hits: %d (in %% %.1f)\n",
				name, total, cacheHit, (float)cacheHit / total * 100, boostCacheHit, (float)boostCacheHit / total * 100);
		long lastTotal = 0;
		long lastCacheHit = 0;
		long lastBoostCacheHit = 0;
		synchronized (semaphore) {
			for (EntryType entryType : last1000) {
				lastTotal++;
				switch (entryType) {
					case BOOST_CACHE_HIT:
						lastBoostCacheHit++;

						break;
					case CACHE_HIT:
						lastCacheHit++;
						break;
					case MISS:
						break;
				}

			}
		}
		String last1000 = String.format("%s statistics for the last 1000 Total: %d Cachehits: %d (in %% %.1f) Boost cache hits: %d (in %% %.1f)",
				name, lastTotal, lastCacheHit, (float)lastCacheHit / lastTotal * 100, lastBoostCacheHit, (float)lastBoostCacheHit / lastTotal * 100);
		return totals + last1000;
	}

	/**
	 * Statistics can be ordered or filtered by group, the group name must not be null
	 *
	 * @return
	 */
	public String getStatisticGroupname() {
		return groupname;
	}
}
