package org.jboss.seam.cache;

import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;
import org.jboss.seam.Component;
import org.jboss.seam.ScopeType;
import org.jboss.seam.annotations.*;
import static org.jboss.seam.annotations.Install.BUILT_IN;
import org.jboss.seam.annotations.intercept.BypassInterceptors;
import org.jboss.seam.contexts.Contexts;
import org.jboss.seam.log.LogProvider;
import org.jboss.seam.log.Logging;
import org.jboss.seam.statistics.StatisticsGenerator;
import org.jboss.seam.statistics.StatisticsManager;


/**
 * Cache to be used as page fragment cache. Is using EhCache as provider and uses
 * <i>org.jboss.seam.cache.DefaultCache</> as the default cache region.
 *
 * @author Sebastian Hennebrueder
 */
@Name("org.jboss.seam.cache.ehcacheprovider")
@Scope(ScopeType.APPLICATION)
@BypassInterceptors
@Install(value = false, precedence = BUILT_IN)
public class EhCacheProvider implements Cache<EhCacheProvider> {

  private CacheManager cacheManager;

  private String cfgResourceName;

  String defaultCacheRegion = "org.jboss.seam.cache.DefaultCache";

  private static final LogProvider log = Logging.getLogProvider(EhCacheProvider.class);

  private boolean enableStatistics = false;

  private StatisticsGenerator statistics;


  /**
   * @return - the instance of the Cache implementation
   */
  public EhCacheProvider getDelegate() {
    return this;
  }

  public void addOrUpdate(String key, Object object) {
    addOrUpdate(defaultCacheRegion, key, object);
  }

  public void addOrUpdate(String region, String key, Object object) {
    net.sf.ehcache.Cache cache = getRegion(defaultCacheRegion);
    Element element = new Element(key, object);
    cache.put(element);
  }

  public void clear() {
    String[] strings = cacheManager.getCacheNames();
    for (String cacheName : strings) {
      net.sf.ehcache.Cache cache = getRegion(cacheName);
      cache.removeAll();
    }
  }


  public Object get(String key) {
    return get(defaultCacheRegion, key);
  }

  public Object get(String region, String key) {
    net.sf.ehcache.Cache cache = getRegion(defaultCacheRegion);
    Element element = cache.get(key);
    if (element != null)
      return element.getObjectValue();
    else
      return null;
  }

  public String getCfgResourceName() {
    return this.cfgResourceName;
  }

  public void remove(String key) {
    remove(defaultCacheRegion, key);
  }

  private net.sf.ehcache.Cache getRegion(String regionName) {
    net.sf.ehcache.Cache result = cacheManager.getCache(regionName);
    if (result == null)
      throw new CacheException("Cache region not found");
    return result;
  }

  public void remove(String region, String key) {
    net.sf.ehcache.Cache cache = getRegion(region);
    cache.remove(key);
  }

  public void setCfgResourceName(String cfgResourceName) {
    this.cfgResourceName = cfgResourceName;
  }

  public void setEnableStatistics(boolean enableStatistics) {
    this.enableStatistics = enableStatistics;
  }

  @Create
  public void start() throws Exception {
    log.debug("starting EhCacheProvider cache");
    // TODO validate if there is any common approach to load resources in Seam
    if (cfgResourceName != null)
      cacheManager = new CacheManager(cfgResourceName);
    else
      cacheManager = new CacheManager();

    if (enableStatistics) {
      statistics = new EhCacheStatistics(cacheManager);
      StatisticsManager.instance().registerStatisticsGenerator(statistics);
    }

  }

  @Destroy
  public void stop() {
    log.debug("stopping EhCacheProvider cache");
    if (enableStatistics) {
      StatisticsManager.instance().unregisterStatisticsGenerator(statistics);
      statistics = null;
    }

    cacheManager.shutdown();
    cacheManager = null;
  }

  public static Cache<EhCacheProvider> instance() {
    if (!Contexts.isApplicationContextActive()) {
      throw new IllegalStateException("No active application scope");
    }
    return (Cache<EhCacheProvider>) Component.getInstance(EhCacheProvider.class, ScopeType.APPLICATION);
  }

}
