package org.jboss.seam.cache;

import org.jboss.seam.Component;
import org.jboss.seam.ScopeType;
import org.jboss.seam.annotations.*;
import static org.jboss.seam.annotations.Install.BUILT_IN;
import org.jboss.seam.annotations.intercept.BypassInterceptors;
import org.jboss.seam.contexts.Contexts;
import org.jboss.seam.log.LogProvider;
import org.jboss.seam.log.Logging;
import org.jboss.seam.statistics.StatisticsManager;

import java.io.File;
import java.io.IOException;

/**
 * Uses the file system as cache. This is useful for dynamic web pages which are expensive to create.
 * This cache has an additional method which returns the timestamp of a file. This allows the
 * to compare the change timestamp of an object with the file timestamp to decide if the file is still up to date.
 * <p/>
 * Configuration:
 * Don't specify a cfgResourceName but only set the root folder where all files are placed in.
 *
 * @author Sebastian Hennebrueder
 */

@Name("org.jboss.seam.cache.filesystemcacheprovider")
@Scope(ScopeType.APPLICATION)
@BypassInterceptors
@Install(value = false, precedence = BUILT_IN)
public class FileSystemCacheProvider implements Cache<FileSystemCacheProvider> {

  private static final LogProvider log = Logging.getLogProvider(FileSystemCacheProvider.class);

  private String rootDirectory;

  public static final String defaultCacheRegion = "org.jboss.seam.cache.DefaultCache";

  private FileSystemCache fileSystemCache;

  private int boostBufferSize = 20;

  private boolean enableStatistics = false;


  /**
   * Directory where cached fragments are saved
   *
   * @param rootDirectory - the root directory
   */
  public void setRootDirectory(String rootDirectory) {
    this.rootDirectory = rootDirectory;
  }

  public FileSystemCacheProvider getDelegate() {
    return null;
  }

  @Create
  public void start() {
    if (rootDirectory == null || rootDirectory.length() == 0) {
      log.debug("No rootDirectory configured using java.io.tmpdir");
      rootDirectory = System.getProperty("java.io.tmpdir");
      rootDirectory = rootDirectory + File.separatorChar + "org.jboss.seam.FileSystemCache";
    }

    // TODO add the context of the application as postfix  when the default temp is used

    log.info("Starting FileSystemCacheProvider with root folder " + rootDirectory);
    File root;
    try {
      root = new File(rootDirectory);
      if (!root.exists())
        root.mkdir();
      else {
        deleteOldCache(root);
      }
    } catch (Exception e) {
      throw new CacheException("Could not create directory for file system cache.", e);
    }

    // test if we can create files
    try {
      File temp = File.createTempFile("test", "txt", root);
      temp.delete();
    } catch (IOException e) {
      throw new CacheException(String.format("Cannot create files for file system cache in %s",
              root.getAbsolutePath()), e);
    }

    fileSystemCache = new FileSystemCache(root, defaultCacheRegion, boostBufferSize);

    if (enableStatistics) {
      CacheStatistics statistics = new CacheStatistics("FileSystemCache");
      fileSystemCache.setCacheStatistics(statistics);
      StatisticsManager.instance().registerStatisticsGenerator(statistics);
    }
  }

  /**
   * deletes empty folders and files ending with .cache
   * @param folder - folder which files and subfolder are deleted
   */
  private void deleteOldCache(File folder) {
    String[] strings = folder.list();
    for (String string : strings) {
      File f = new File(folder + File.separator + string);
      if (f.isDirectory()) {
        deleteOldCache(f);
        f.delete();
      } else if (string.endsWith(FileSystemCache.SUFFIX))
        f.delete();

      
    }
  }

  @Destroy
  public void stop() {
    log.info("Stopping FileSystemCacheProvider cache with root folder " + rootDirectory);
    if (enableStatistics) {
      CacheStatistics statistics = fileSystemCache.getCacheStatistics();
      StatisticsManager.instance().unregisterStatisticsGenerator(statistics);
      fileSystemCache.setCacheStatistics(null);
      fileSystemCache = null;
    }
  }

  public static Cache<FileSystemCacheProvider> instance() {
    if (!Contexts.isApplicationContextActive()) {
      throw new IllegalStateException("No active application scope");
    }
    return (Cache<FileSystemCacheProvider>) Component.getInstance(FileSystemCacheProvider.class, ScopeType.APPLICATION);
  }

  public void setEnableStatistics(boolean enableStatistics) {
    this.enableStatistics = enableStatistics;
  }

  public boolean isEnableStatistics() {
    return enableStatistics;
  }

  public String getCfgResourceName() {
    return null;
  }

  public void setCfgResourceName(String cfgResourceName) {

  }

  public void setBoostBufferSize(int boostBufferSize) {
    this.boostBufferSize = boostBufferSize;
  }

  public Object get(String key) {
    return fileSystemCache.readFile(key);
  }


  public Object get(String region, String key) {
    return fileSystemCache.readFile(region, key);
  }

  public void addOrUpdate(String key, Object object) {
    fileSystemCache.createOrUpdateFile(key, (String) object);
  }

  public void addOrUpdate(String region, String key, Object object) {
    fileSystemCache.createOrUpdateFile(region, key, (String) object);
  }

  public void remove(String key) {
    fileSystemCache.removeFile(key);
  }

  public void remove(String region, String key) {
    fileSystemCache.removeFile(region, key);
  }

  public void clear() {
    fileSystemCache.clear();
  }

}
