package org.jboss.seam.cache;

import org.jboss.cache.aop.PojoCache;
import org.jboss.cache.*;
import org.jboss.cache.CacheException;
import org.jboss.seam.log.LogProvider;
import org.jboss.seam.log.Logging;
import org.jboss.seam.annotations.*;
import static org.jboss.seam.annotations.Install.BUILT_IN;
import org.jboss.seam.annotations.intercept.BypassInterceptors;
import org.jboss.seam.ScopeType;
import org.jboss.seam.Component;
import org.jboss.seam.statistics.StatisticsGenerator;
import org.jboss.seam.statistics.StatisticsManager;
import org.jboss.seam.contexts.Contexts;

/**
 * JBoss Treecache is suitable for simple objects which do not require a per field synchronization like
 * the JBoss Pojo Cache. It requires less memory and is faster as the PojoCache.
 * <p/>
 * This version of the cache provider is suitable for JBoss Cache 1.x only!!!
 *
 * @author Sebastian Hennebrueder
 */

@Name("org.jboss.seam.cache.jbosstreecacheprovider")
@Scope(ScopeType.APPLICATION)
@BypassInterceptors
@Install(value = false, precedence = BUILT_IN)
public class JbossTreeCacheProvider implements Cache<TreeCache> {

  private TreeCache treeCache;

  private CacheStatistics statisticsGenerator;

  private boolean enableStatistics = false;

  private static final LogProvider log = Logging.getLogProvider(JbossTreeCacheProvider.class);

  public static final String defaultCacheRegion = "org.jboss.seam.cache.DefaultCache";
  public static final String pageFragmentTree = "org.jboss.seam.cache.PageFragmentCache";

  private String cfgResourceName;

  private void writeStatistic(CacheStatistics.EntryType entryType) {
    if (enableStatistics)
      statisticsGenerator.addEntry(entryType);
  }

  private Fqn buildFqn(String region) {
    if (region != null)
      return new Fqn(new String[]{pageFragmentTree, region});
    else
      return new Fqn(new String[]{pageFragmentTree});
  }

  @Create
  public void start() throws org.jboss.seam.cache.CacheException {
    log.info("starting JBoss tree cache");

    try {
      treeCache = new TreeCache();
      new PropertyConfigurator().configure(treeCache, cfgResourceName);

      treeCache.createService();
      treeCache.startService();
      if (enableStatistics) {
        statisticsGenerator = new CacheStatistics("JBossTreeCache");
        StatisticsManager.instance().registerStatisticsGenerator(statisticsGenerator);
      }
    } catch (Exception e) {
      throw new org.jboss.seam.cache.CacheException("Error starting JBoss tree cache", e);
    }
  }

  @Destroy
  public void stop() {
    log.info("stopping JBoss tree cache");
    if (enableStatistics) {
      StatisticsManager.instance().unregisterStatisticsGenerator(statisticsGenerator);
      statisticsGenerator.reset();
      statisticsGenerator = null;
    }
    treeCache.stopService();
    treeCache.destroyService();
    treeCache = null;
  }

  public static Cache<JbossTreeCacheProvider> instance() {
    if (!Contexts.isApplicationContextActive()) {
      throw new IllegalStateException("No active application scope");
    }
    return (Cache<JbossTreeCacheProvider>) Component.getInstance(JbossTreeCacheProvider.class, ScopeType.APPLICATION);
  }

  public String getCfgResourceName() {
    return cfgResourceName;
  }

  public void setCfgResourceName(String cfgResourceName) {
    this.cfgResourceName = cfgResourceName;
  }

  public Object get(String key) {
    return get(defaultCacheRegion, key);
  }


  public Object get(String region, String key) {
    try {
      Node node = treeCache.get(buildFqn(region));
      String content = null;
      if (node != null)
        content = (String) node.get(key);
      if (content == null)
        writeStatistic(CacheStatistics.EntryType.MISS);
      else
        writeStatistic(CacheStatistics.EntryType.CACHE_HIT);
      return content;
    } catch (CacheException e) {
      throw new org.jboss.seam.cache.CacheException(
              String.format("Cache throw exception when trying to get %s from region %s.", key, region), e);
    }
  }

  public void addOrUpdate(String key, Object object) {
    addOrUpdate(defaultCacheRegion, key, object);
  }

  public void addOrUpdate(String region, String key, Object object) {
    try {
      treeCache.put(buildFqn(region), key,
              object);
    } catch (CacheException e) {
      throw new org.jboss.seam.cache.CacheException(
              String.format("JBoss Cache throw exception when adding object for key %s to region %s", key, region),e);
    }
  }


  public void remove(String key) {
    remove(defaultCacheRegion, key);
  }

  public void remove(String region, String key) {
    try {
      treeCache.remove(buildFqn(region), key);
    } catch (CacheException e) {
      throw new org.jboss.seam.cache.CacheException(
              String.format("JBoss Cache throw exception when removing object for key %s in region %s", key, region),e);
    }

  }

  public void clear() {
    try {
      treeCache.remove(buildFqn(null));
    } catch (CacheException e) {
      throw new org.jboss.seam.cache.CacheException(
              String.format("JBoss Cache throw exception when clearing default cache."),e);
    }
  }

  /**
   * @return - the instance of the Cache implementation
   */
  public TreeCache getDelegate() {
    return treeCache;
  }

  public boolean isEnableStatistics() {
    return enableStatistics;
  }

  public void setEnableStatistics(boolean enableStatistics) {
    this.enableStatistics = enableStatistics;
  }
}